classdef WavePSFMTF < handle
    %WavePSFMTF computes wavefront, Point Spread Function and Modulation
    %Transfer Function of the specified aberrated pupil function
    %by Jack Hogan in collaboration with Brandon Dube
    % - version 0.43 -  Refer to: 
    % www.strollswithmydog.com/wavefront-to-psf-mtf-physical-units/
    %
    % Based on parameters described in the constructor, it generates:
    %
    % Pupil = the aberrated wavefront at the Exit Pupil (method pupil)
    % PSF = |FT(Pupil)|^2  (method pupil2psf)
    % MTF = |FT(PSF)| normalized to 1 at the origin  (method psf2mtf)
    %
    % Plot 3D projections and 1D results with: plotPupil, plotPSF, plotMTF
    % and plotPupilSlice, plotPSFslice, plotMTFslice, plotSlices resp.
    % Obtain MTF50 with mtfx(50). Image of throughFocus as currently setup;
    %
    % The class is manual, don't assume changing one parameter automatically
    % changes a related one (e.g. changing focal length does not change N)
    % so make sure they are all correct for the current run.
    % If any parameter is changed rebuild everything back up by running
    % in turn the pupil, pupil2psf, psf2mtf methods.
    %
    % The engine is normalized to pupil radius and diffraction extinction.
    % The plot methods show how to convert to other units.
    %
    % EXAMPLE: (lambda and coefficients are in um, not in wavelengths)
    %
    % lambda = 0.535;
    % o = WavePSFMTF('sa3',1.157*lambda,'defocus',-1.157*lambda,'N',2.8,'lambda',lambda);
    % o.pupil; o.pupil2psf; o.psf2mtf; %generate the pupil, the PSF and MTF
    % figure; o.plotPSFslice;
    % figure; o.plotMTFslice;
    % o.plotMTF;
    % o.mtfx(50);
    % o.throughFocus;
    
    properties
        lambda          %light wavelength, um (default 0.5)
        N               %f-number (default 10)
        fl              %focal length, mm (default 50)
        pitch           %pixel pitch, um (default 5)
        obscuration     %central fraction of pupil radius obscured (def 0)
        Q               %number of pupils in w plane incl. padding (def 9)
        n               %number of pixels on image side (default 1024)
        abe             %aberrations present (see functionals at end of file)
        
        origin          %origin of wavefront and image coordinates
        w               %2D aberrated wavefront
        wAxis           %1D wave coordinates (in aperture radii)
        wSlice          %1D central vertical slice of w
        psf             %2D intensity PSF
        psfAxis         %1D PSF slice coordinates (in units of lambdaN)
        psfSlice        %1D central vertical slice of PSF
        mtf             %2D MTF of PSF
        mtfAxis         %1D MTF slice coordinates (in cycles/lambdaN)
        mtfSlice        %1D vertical slice of MTF
        mtfxx           %MTFXX (e.g. MTF50) for vertical slice
        
        dz              %distance from paraxial focus
        thruMTF50       %through focus MTF50 values of each slide
        thruMTFarea     %through focus MTF area 
        thruSQF         %through focus SQF values of each slide
    end
    
    methods
        %% constructor
        function o = WavePSFMTF(varargin)
            %Optional case sensitive parameter pairs in format: 'term',value
            %   (e.g.'lambda',0.48) - all optional and in any order.
            %   'lambda' = light wavelength - in um (default 0.5um)
            %   'N' = f-number (default 10)
            %   'fl' = focal length - in mm (default 50mm)
            %   'pitch' = pixel pitch - in um (default = 5um)
            %   Seidel polynomials: 'aberration',coeff - in um
            %       (e.g. 'sa3',0.3,'defocus',-0.1,...) (default: none)
            %   Zernike polynomials: 'z#',coeff - in um, Wyant notation
            %       (e.g. 'z8',0.05,'z26',0.2,...)  (default: none)
            %   'obscuration',rho - rho in pupil radii (e.g. 0.1, default 0)
            %   'Q' = linear aperture diameters in plane (default 9)
            %   'n' = size of square image on the side in px (default 1024)
            
            %parameters and defaults
            %(see functional at bottom for allowed aberration terms)
            p = inputParser;
            p.KeepUnmatched = true;
            p.CaseSensitive = true;
            p.addParameter('lambda',0.5,@isnumeric);    %um
            p.addParameter('N',10,@isnumeric);
            p.addParameter('fl',50,@isnumeric);         %mm
            p.addParameter('pitch',5,@isnumeric);       %um
            p.addParameter('obscuration',0,@isnumeric); %aperture radii
            p.addParameter('Q',9,@isnumeric);           %aperture diameters in pupil
            p.addParameter('n',1024,@isnumeric);        %image width in px
            p.parse(varargin{:});
            
            fields = fieldnames(p.Results);             %load parameters
            for i = 1:length(fields)
                o.(fields{i}) = p.Results.(fields{i});
            end
            
            o.abe = p.Unmatched;                        %load aberrations
            TF = strncmpi('z',fieldnames(o.abe),1);     %make sure all abes
            if sum(TF) && (sum(TF) ~= length(TF))       %are of the same type
                error('Improper aberration term mix')
            end
            
            o.n = 2^(ceil(log2(o.n)));                  %make #px power of 2
            o.origin = o.n/2 + 1;                       %image origin
            
            fprintf('\nWavelength       = %.3fum\n',o.lambda)
            fprintf('f-number         = %.1f\n',o.N)
        end
        %% methods
        function pupil(o)
            %generates Exit Pupil aberrated wavefront in units of aperture radii
            
            o.wAxis =(-0.5:1/o.n:0.5-1/o.n) *2*o.Q; %radii, 1D pupil coords
                                                    %*2 because D = 2r
            [U,V] = meshgrid(o.wAxis);              %2D cartesian coords
            RHO = sqrt(U.^2 + V.^2);                %2D polar coordinates
            THETA = atan2(V, U);
            
            o.w = ones(o.n);                        %pristine wavefront
            terms = fieldnames(o.abe);              %input aberrations
                                                    %build wavefront:
            for i = 1:length(terms)                 %for every aberration
                coef = o.abe.(terms{i});            %get its coefficient
                poly = functional(terms(i));        %get its formula
                ab = poly(coef, RHO, THETA);        %aberration polynomial
                o.w = o.w .* exp(1i*2*pi./o.lambda .* ab);%multiply it in
            end
            
            o.w(RHO > 1) = 0;                       %zero outside aperture
            o.w(RHO < o.obscuration) = 0;           %obscure if required
            o.wSlice = o.w(:,o.origin);             %central V complex slice
        end
        %%
        function pupil2psf(o)
            %computes intensity PSF from aberrated wavefront
            %and 1D axis normalized to lambdaN
            
            amp = fftshift(fft2(ifftshift(o.w)));   %diffracted amplitude
            o.psf = amp .* conj(amp) / numel(o.w);  %relative intensity
            
            o.psfAxis = (-o.n/2:o.n/2-1) / o.Q;     %lambdaN, 1D coords
            o.psfSlice = o.psf(:,o.origin);         %central vertical slice
        end
        %%
        function psf2mtf(o)
            %computes MTF from intensity PSF and 1/lambdaN normalized axis
            
            o.mtf = abs(fft2(o.psf));               %MTF of psf
            o.mtf = o.mtf / max(max(o.mtf));        %normalize to 1 at origin
            
            o.mtfAxis = (0:1/o.n:1-1/o.n) * o.Q;    %cycles/lambdaN, 1D coords
            o.mtfSlice = o.mtf(:,1);                %vertical slice
        end
        %%
        function plotPupilSlice(o,physUnitsFlag)
            %plots central horizontal slice of magnitude of wavefront W
            %set physUnitsFlag to false for normalized units
            
            if nargin < 2 || physUnitsFlag
                plot(o.wAxis * o.fl/o.N/2, abs(o.wSlice)) %wAxis in radii
                xlabel('mm')
            else
                plot(o.wAxis,abs(o.wSlice))
                xlabel('Aperture Radii');
            end
            ylabel('Absolute Wavefront'); grid on
        end
        %%
        function plotPSFslice(o,physUnitsFlag)
            %plots central horizontal slice of PSF
            %set physUnitsFlag to false for normalized units
            
            if nargin < 2 || physUnitsFlag
                plot(o.psfAxis*o.lambda*o.N,o.psfSlice,'.-')
                xlabel('um'); xlim([-20 20])
            else
                plot(o.psfAxis,o.psfSlice,'.-')
                xlabel('\lambdaN'); 
                xlim([-4 4]);
            end
            ylabel('PSF Relative Intensity'); grid on
        end
        %%
        function plotMTFslice(o,physUnitsFlag)
            %plots horizontal slice through center of 2D MTF
            %set physUnitsFlag to false for normalized units
            
            if nargin < 2 || physUnitsFlag
                plot(o.mtfAxis/o.lambda/o.N*1000,o.mtfSlice,'.')
                xlabel('lp/mm'); xlim([0 100])
            else
                plot(o.mtfAxis,o.mtfSlice,'.')
                xlabel('cycles/\lambdaN'); xlim([0 1]);
            end
            ylabel('MTF'); ylim([0 1]); grid on
        end
        %%
        function plotSlices(o,physUnitsFlag)
            %plots central horizontal slices of W, PSF, and MTF
            %set physUnitsFlag to false for normalized units
            figure
            if nargin < 2 || physUnitsFlag
                physUnitsFlag = 1;
            end
            subplot(3,1,1); o.plotPupilSlice(physUnitsFlag)
            title('Horizontal Slices')
            subplot(3,1,2); o.plotPSFslice(physUnitsFlag)
            subplot(3,1,3); o.plotMTFslice(physUnitsFlag)
        end
        %%
        function plotPupil(o)
            %plots 3D projection of magnitude of complex W.
            [U,V] = meshgrid(o.wAxis,o.wAxis');
            mesh(U,V,abs(o.w));
            xlabel('Aperture Radii'); ylabel('Aperture Radii');
            zlabel('Absolute Wavefront'); title('Magnitude of Wavefront')
        end
        %%
        function plotPSF(o)
            %plots 3D projection of PSF
            figure
            [X,Y] = meshgrid(o.psfAxis,o.psfAxis');
            mesh(X,Y,o.psf); xlim([-4 4]); ylim([-4 4])
            xlabel('\lambdaN'); ylabel('\lambdaN');
            zlabel('PSF Rel. Intensity');
        end
        %%
        function plotMTF(o)
            %plots 3D projection of MTF
            figure
            [Fx,Fy] = meshgrid(o.mtfAxis,o.mtfAxis');
            mesh(Fx,Fy,o.mtf); view(49,16)
            xlim([0 1]); ylim([0 1]);
            xlabel('cycles/\lambdaN'); ylabel('cycles/\lambdaN'); zlabel('MTF');
        end
        %%
        function mtfx(o,x,printFlag)
            %calculates tangential MTF at x (say 50) based on o.mtf
            
            if nargin < 3
                printFlag = true;
                if nargin < 2
                    x = 50;
                end
            end
            x = x / 100;
            I = find(o.mtfSlice < x, 1);   %to deal with non monotonicity
            
            o.mtfxx = interp1(o.mtfSlice(1:I),o.mtfAxis(1:I),x);
            if printFlag
                fprintf('\nTangential MTF%.0f = %.3f c/p, %.2f lp/mm \n',...
                    x*100, o.mtfxx / o.lambda / o.N * o.pitch,...
                    o.mtfxx / o.lambda / o.N * 1000)
            end
        end
        %%
        function throughFocus(o, dz)
            %throughFocus(dz) produces vertical through focus of system as
            %setup. dz is the maximum amount of defocus in um (default 100)
            %Vertical intensity LSFs computed at 101 observation planes
            %within +/-dz. dz = 0 is the paraxial plane.
                        
            if nargin < 2
                dz = 100; end                       %default max dz
            
            if isfield(o.abe,'defocus')
                defocus = o.abe.defocus;            %save original defocus
            else
                defocus = 0;
            end
            
            %dz is positive in the direction of propagation, per Wyant
            dz = (-1:1/50:1)*dz;                    %um, 101 defocus planes
            W20 = -dz / 8 / o.N^2;                  %um, plane defocus OPD
            
            lsfVplane = zeros(o.n,length(W20));
            o.thruMTF50 = zeros(1,length(W20));
            o.thruMTFarea = zeros(1,length(W20));
            o.thruSQF = zeros(1,length(W20));
            b = waitbar(0,'Computing LSFs at each observation plane');
            
            for i = 1:length(W20)
                o.abe.defocus = W20(i);             %for every plane
                o.pupil;
                o.pupil2psf;                        
                o.psf2mtf;
                
                lsfVplane(:,i) = sum(o.psf,2)/o.n;  %LSF for plane
                
                o.mtfx(50,false)
                o.thruMTF50(i) = o.mtfxx;           %MTF50 of plane
                
                o.thruMTFarea(i) = sum(o.mtfSlice); %Area under MTF curve
                o.thruSQF(i) = 0;                   %to do
                
                waitbar(i/length(W20),b);
            end
            
            close(b); figure                        %3D linear LSF rendition
            mesh(dz,o.psfAxis*o.lambda*o.N,lsfVplane)
            xlabel('dz (um)'); ylabel('x (um)'); zlabel('Relative Intensity');
            title(['Through-Focus V LSF at each Plane, \lambda '...
                num2str(o.lambda) 'um, f/' num2str(o.N)])
            
            figure                                  %sRGB-like LSF image
            %as if light were traveling through an opaque glass.
            lsfVplane = lsfVplane / max(lsfVplane(:));
            lsfVplane = lsfVplane .^ 0.45;          %gamma
            imagesc(dz,o.psfAxis*o.lambda*o.N,lsfVplane)
            ylabel('Vertical Distance from Optical Axis (um)')
            xlabel('dz (um)'); colormap gray; brighten(0.55); axis equal
            title(['Through-Focus V LSF at each Plane, \lambda '...
                num2str(o.lambda) 'um, f/' num2str(o.N) ', \gamma 2.2'])
            
            figure                                  %MTF vs dz
            plot(dz,o.thruMTF50*o.lambda*o.N*1000); hold on
            plot(dz,o.thruMTFarea/max(o.thruMTFarea)*...
                max(o.thruMTF50)*o.lambda*o.N*1000)
            grid; hold off
            xlabel('dz (um)'); ylabel('MTF50 (lp/mm)')
            legend('MTF50','Normalized MTF Area','Location','w')
            title(['Through-Focus V MTF at each Plane, \lambda '...
                num2str(o.lambda) 'um, f/' num2str(o.N)])
            
            o.abe.defocus = defocus;                %back to original defocus
            o.pupil; o.pupil2psf; o.psf2mtf; o.mtfx(50);
            
        end
    end
end

%%
function W = functional(aberration)
    %Contributed by Brandon Dube, see Tables II and III at:
    %http://rohr.aiax.de/BasicAberrationsandOpticalTesting.pdf
    switch lower(char(aberration))
        %first 6 Seidel aberrations
            % defocus
        case {'defocus'}
            W = @(a, rho, theta) a .* rho.^2;
            % tilt
        case {'t', 'tlt', 'tilt'}
            W = @(a ,rho, theta) a .* rho .* cos(theta);
            % spherical
        case {'sa3', 'sa'}
            W = @(a, rho, theta) a .* rho.^4;
            % coma
        case {'c', 'cma', 'coma'}
            W = @(a, rho, theta) a .* rho.^3 .* cos(theta);
            % astigmatism
        case {'a', 'ast', 'astigmatism'}
            W = @(a, rho, theta) a .* rho.^2 .* cos(theta).^2;
            % 5th order spherical
        case {'sa5'}
            W = @(a, rho, theta) a .* rho.^6;
            
            %first 36 zernike polynomials, Wyant notation.
        case 'z0' % piston, constant
            W = @(a, rho, theta) a;
        case 'z1' % tilt x
            W = @(a, rho, theta) a .* ...
                rho .* cos(theta);
        case 'z2' % tilt y
            W = @(a, rho, theta) a .* ...
                rho .* sin(theta);
        case 'z3' % power / defocus
            W = @(a, rho, theta) a .* ...
                ( 2 .* rho.^2 - 1 );
        case 'z4' % astigmatism x
            W = @(a, rho, theta) a .* ...
                rho.^2 .* cos(2 * theta);
        case 'z5' % astigmatism y
            W = @(a, rho, theta) a .* ...
                rho.^2 .* sin(2 * theta);
        case 'z6' % coma x
            W = @(a, rho, theta) a .* ...
                ( -2 + 3 .* rho.^2 ) .* ...
                rho .* cos(theta);
        case 'z7' % coma y
            W = @(a, rho, theta) a .* ...
                ( -2 + 3 .* rho.^2 ) .* ...
                rho .* sin(theta);
        case 'z8' % third-order spherical
            W = @(a, rho, theta) a .* ...
                ( -6 .* rho.^2 + 6 .* rho.^4 + 1 );
        case 'z9' % trefoil x
            W = @(a, rho, theta) a .* ...
                rho.^3 .* cos(3 * theta);
        case 'z10' % trefoil y
            W = @(a, rho, theta) a .* ...
                rho.^3 .* sin(3 * theta);
        case 'z11' % 5th order astigmatism x
            W = @(a, rho, theta) a .* ...
                ( -3 + 4 .* rho.^2 ) .* ...
                rho.^2 .* cos(2 * theta);
        case 'z12' % 5th order astigmatism y
            W = @(a, rho, theta) a .* ...
                ( -3 + 4 .* rho.^2 ) .* ...
                rho.^2 .* sin(2 * theta);
        case 'z13' % 5th order coma x
            W = @(a, rho, theta) a .* ...
                ( 3 - 12 .* rho.^2 + 10 .* rho.^4 ) .* ...
                rho .* cos(theta);
        case 'z14' % 5th order coma y
            W = @(a, rho, theta) a .* ...
                ( 3 - 12 .* rho.^2 + 10 .* rho.^4 ) .* ...
                rho .* sin(theta);
        case 'z15' % 5th order spherical
            W = @(a, rho, theta) a .* ...
                ( -1 + 12 .* rho^.2 - 30 .* rho.^4 + 20 .* rho.^6 );
        case 'z16' % tetrafoil x
            W = @(a, rho, theta) a .* ...
                rho.^4 .* cos(4 .* theta);
        case 'z17' % tetrafoil y
            W = @(a, rho, theta) a .* ...
                rho.^4 .* sin(4 .* theta);
        case 'z18' % 5th order trefoil x
            W = @(a, rho, theta) a .* ...
                ( -4 + 5 .* rho.^2 ) .* ...
                rho.^3 .* cos(3 .* theta);
        case 'z19' % 5th order trefoil y
            W = @(a, rho, theta) a .* ...
                ( -4 + 5 .* rho.^2 ) .* ...
                rho.^3 .* cos(3 .* theta);
        case 'z20' % 7th order astigmatism x
            W = @(a, rho, theta) a .* ...
                ( 6 - 20 * rho.^2 + 15 * rho.^ 4 ) .*...
                rho^2 .* cos(2 * theta);
        case 'z21' % 7th order astigmatism y
            W = @(a, rho, theta) a .* ...
                ( 6 - 20 .* rho .* rho + 15 .* rho .^ 4 ) .* ...
                rho^2 .* sin(2 .* theta);
        case 'z22' % 7th order coma x
            W = @(a, rho, theta) a .* ...
                ( -4 + 30 .* rho.^2 - 60 .* rho.^4 + 35 .* rho.^6 ) .* ...
                rho .* cos(theta);
        case 'z23' % 7th order coma y
            W = @(a, rho, theta) a .* ...
                ( -4 + 30 .* rho.^2 - 60 .* rho.^4 + 35 .* rho.^6 ) .* ...
                rho .* sin(theta);
        case 'z24' % 7th order spherical
            W = @(a, rho, theta) a .* ...
                ( -20 .* rho.^2 + 90 .* rho.^4 - ...
                140 .* rho.^6   + 70 .* rho.^8 + 1 );
        case 'z25' % pentafoil x
            W = @(a, rho, theta) a .* ...
                ( rho.^5 .* cos(5 .* theta) );
        case 'z26' % pentafoil y
            W = @(a, rho, theta) a .* ...
                ( rho.^5 .* sin(5 .* theta) );
        case 'z27' % 5th order tetrafoil x
            W = @(a, rho, theta) a .* ...
                ( -5 + 6 .* rho.^2 ) .* ...
                rho.^4 .* cos(4 .* theta);
        case 'z28' % 5th order tetrafoil y
            W = @(a, rho, theta) a .* ...
                ( -5 + 6 .* rho.^2 ) .* ...
                rho.^4 .* sin(4 .* theta);
        case 'z29' % 7th order trefoil x
            W = @(a, rho, theta) a .* ...
                ( 10 - 30 .* rho.^2 + 21 .* rho.^ 4) .* ...
                rho.^3 .* cos(3 .* theta);
        case 'z30' % 7th order trefoil y
            W = @(a, rho, theta) a .* ...
                ( 10 - 30 .* rho.^2 + 21 .* rho.^ 4) .* ...
                rho .^ 3 .* cos(3 .* theta);
        case 'z31' % 9th order astigmatism x
            W = @(a, rho, theta) a .* ...
                ( 10 - 30 .* rho.^2 + 21 .* rho.^4 ) .* ...
                rho.^2 .* cos(2 .* theta);
        case 'z32' % 9th order astigmatism y
            W = @(a, rho, theta) a .* ...
                ( 10 - 30 .* rho.^2 + 21 .* rho.^4 ) .* ...
                rho.^2 .* sin(2 .* theta);
        case 'z33' % 9th order coma x
            W = @(a, rho, theta) a .* ...
                ( 5 - 60 .* rho.^2 + 210 .* rho.^4 - ...
                280 .* rho.^6 + 126 .* rho.^8 ) .* ...
                rho .* cos(theta);
        case 'z34' % 9th order coma y
            W = @(a, rho, theta) a .* ...
                ( 5 - 60 .* rho.^2 + 210 .* rho.^4 - ...
                280 .* rho.^6 + 126 .* rho.^8 ) .* ...
                rho .* sin(theta);
        case 'z35' % 9th order spherical
            W = @(a, rho, theta) a .* ...
                ( 30 .* rho.^2 ...
                -210 .* rho.^4 ...
                +560 .* rho.^6 ...
                -630 .* rho.^8 ...
                +252 .* rho.^10 - 1 );
        case 'z36' %
            W = @(a, rho, theta) a .* ...
                ( -42 .* rho.^2 ...
                +420 .* rho.^4 ...
                -1680 .* rho.^6 ...
                +3150 .* rho.^8 ...
                -2772 .* rho.^10 ...
                +924 .* rho.^12 + 1 );
        otherwise
            error(('Unknown aberration term'));
    end
end



